#!/usr/local/bin/tclsh
#-----------------------------------------------------------------------------
#
#	rfc2mib.tcl
#
#	Command-line script to strip RFC headers/docs from MIB modules.  This
#	script should be smart enough to recognize ASN.1-style comments before
#	or intermixed within the module header, as well as match a module header
#	even if it is broken across several lines.
#
#	Copyright (C)2003 Muonics, All Rights Reserved.
#
#	Muonics hereby grants permission to use, copy, modify, distribute,
#	and license this software and its documentation for any purpose, provided
#	that existing copyright notices are retained in all copies and that this
#	notice is included verbatim in any distributions. No written agreement,
#	license, or royalty fee is required for any of the authorized uses.
#	Modifications to this software may be copyrighted by their authors
#	and need not follow the licensing terms described here, provided that
#	the new terms are clearly indicated on the first page of each file where
#	they apply.
#
#	IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
#	FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
#	ARISING OUT OF THE USE OF THIS FILE, ITS DOCUMENTATION, OR ANY
#	DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
#	POSSIBILITY OF SUCH DAMAGE.
#
#	THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
#	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS FILE
#	IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
#	NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
#	MODIFICATIONS.
#
#	GOVERNMENT USE: If you are acquiring this software on behalf of the
#	U.S. government, the Government shall have only "Restricted Rights"
#	in the software and related documentation as defined in the Federal
#	Acquisition Regulations (FARs) in Clause 52.227.19 (c) (2).  If you
#	are acquiring the software on behalf of the Department of Defense, the
#	software shall be classified as "Commercial Computer Software" and the
#	Government shall have only "Restricted Rights" as defined in Clause
#	252.227-7013 (c) (1) of DFARs.  Notwithstanding the foregoing, the
#	authors grant the U.S. Government and others acting in its behalf
#	permission to use and distribute this software in accordance with the
#	terms specified in this license.
#
#-----------------------------------------------------------------------------

namespace eval RFC2MIB {
	# Regular expressions used by the module extraction routines

	variable mRxPageBreak	\
		{[^\n]*\n+\f\n+[^\n]*}

	variable mRxComment \
		"--((\[^-\\n\]*)|(-\[^-\\n\]+))*((--)|\\n)"

	variable mRxWS \
		"(\\s|($mRxComment))"

	variable mRxUCID \
		"\[A-Z\](-?\[A-Za-z0-9\]+)*"

	variable mRxID \
		"(\[A-Za-z\](-?\[A-Za-z0-9\]+)*)"

	variable mRxTagDefaults \
		"(((EXPLICIT)|(IMPLICIT))($mRxWS)+TAGS($mRxWS)+)"

	variable mRxModuleHdr \
		"($mRxWS)*($mRxUCID)($mRxWS)+(PIB-)?DEFINITIONS($mRxWS)+($mRxTagDefaults)?::=($mRxWS)+BEGIN($mRxWS)+"

	variable mRxMacroHdr \
		"($mRxWS)*($mRxUCID)($mRxWS)+MACRO($mRxWS)+::=($mRxWS)+BEGIN($mRxWS)*"

	variable mRxEnd \
		"($mRxWS)*END($mRxWS)*"

	variable mRxSyms \
		{[()-.{}:=<>[\];|]}
}

proc RFC2MIB::Convert { } {
	variable mRxPageBreak
	variable mRxModuleHdr
	variable mRxMacroHdr
	variable mRxEnd
	variable mRxWS
	variable mRxID
	variable mRxUCID
	variable mRxSyms
	global argv

	set channel stdout
	set split 0

	if {[lindex $argv 0] == "-s"} {
		set split 1
		set argv [lrange $argv 1 end]
	}

	switch -- [llength $argv] {
		1 {}

		2 { set channel [open [lindex $argv 1] w] }

		default {
			puts "Usage:"
			puts ""
			puts "  rfc2mib \[-s\] infile \[outfile\]"
			puts ""
			puts "Options:"
			puts ""
			puts "  -s  split into separate files for each module."
			puts ""
			exit -1
		}
	}

	# Read text file contents

	set hFileIn	[open [lindex $argv 0] "r"]
	set data	[read $hFileIn]

	close $hFileIn

	# Remove all carriage returns from the input data

	regsub -all [format %c 13] $data "" data

	# Extract headers/footers from the document

	set first 0

	while {[regexp -start $first -indices $mRxPageBreak $data index]} {
		set first	[lindex $index 0]
		set last	[lindex $index 1]

		set data [string replace $data $first $last "\n\n"]
	}

	# Replace all occurances of 3 or more newlines with two newlines
	# (ie., at most one blank line between paragraphs/sections/etc.)

	regsub -all {\n{3,}} $data "\n\n" data

	# Extract MIB Modules from the surrounding documentation.
	# Proceed until there are no more lines left.

	set start	0
	set end		0
	set length	[string length $data]

	while {$end < $length} {
		# Find the next module header, including any leading whitespace
		# and/or ASN.1 style comments.

		if {![regexp -start $start -indices -- $mRxModuleHdr $data rng]} {
			# No module header matched - all done.
			break
		}

		set start	[lindex $rng 0]
		set end		[lindex $rng 1]

		if {($channel != "stdout") || $split} {
			set module "UNKNOWN"
			set wsstart $start

			while {[regexp -start $wsstart -indices -- $mRxWS $data rng]} {
				set wsstart [expr {[lindex $rng 1] + 1}]
			}
			regexp -start $start $mRxUCID $data module
			puts $module
		}

		set depth 1

		while {$depth && ($end < $length)} {
			incr end

			# Match up any whitespace, quoted strings, symbols or numbers.
			# Doing strings before identifier is important to prevent matching
			# "END" or "MACRO" inside a quoted string.

			if {[regexp -start $end -indices -- "\\A($mRxWS)+" $data rng] ||
				[regexp -start $end -indices -- {\A"[^"]*"} $data rng] ||
				[regexp -start $end -indices -- "\\A($mRxSyms)+" $data rng] ||
				[regexp -start $end -indices -- {\A[[:digit:]]+} $data rng]} {

				set end [lindex $rng 1]
				continue
			}

			# The only thing left as a likely match would be an identifier.

			if {[regexp -start $end -indices -- "\\A($mRxID)\\s*" $data rng]} {
				foreach { idStart end } $rng break

				# Check for the END or MACRO keywords, and adjust the number
				# of required END keywords accordingly.  Done extracting
				# this module if no more END keywords are required.

				switch -- [string trim [string range $data $idStart $end]] {
					"END" {
						incr depth -1
					}

					"MACRO" {
						incr depth
					}
				}

				continue
			}
		}

		if {$split} {
			set spchan [open $module.mib "w"]
			puts $spchan [string range $data $start $end]
			close $spchan
		}

		if {($channel != "stdout") || !$split} {
			puts $channel [string range $data $start $end]
		}

		set start [expr {$end + 1}]
	}

	if {$channel != "stdout"} {
		close $channel
	}
}

RFC2MIB::Convert

